#include <sys/param.h>
#include <sys/types.h>
#include <sys/devctl.h>
#include <sys/errno.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <netlink/netlink.h>
#include <netlink/netlink_ctl.h>
#include <netlink/netlink_var.h>
#include <netlink/netlink_generic.h>

#include <netlink_sysevent.h>

#define	CTRL_FAMILY_NAME	"nlsysevent"
static uint32_t ctrl_family_id;
static uint32_t all_group_id;

static struct sysevent {
	uint32_t id;
	const char *name;
} sysevents [] = {
	{ 0, "ACPI"},
	{ 0, "AEON"},
	{ 0, "CAM"},
	{ 0, "CARP"},
	{ 0, "coretemp"},
	{ 0, "DEVFS"},
	{ 0, "ETHERNET"},
	{ 0, "GEOM"},
	{ 0, "HYPERV_NIC_VF"},
	{ 0, "IFNET"},
	{ 0, "INFINIBAND"},
	{ 0, "kern"},
	{ 0, "kernel"},
	{ 0, "nvme"},
	{ 0, "PMU"},
	{ 0, "RCTL"},
	{ 0, "USB"},
	{ 0, "VFS"},
	{ 0, "VT"},
	{ 0, "ZFS"},
};

static void
sysevent_send(const char *system, const char *subsystem, const char *type,
    const char *data)
{
	struct nl_writer nw = {};
	uint32_t group_id = 0;

	for (size_t i = 0; nitems(sysevents); i++) {
		if (strcmp(sysevents[i].name, system) == 0) {
			group_id = sysevents[i].id;
			break;
		}
	}

	if (__predict_false(group_id == 0)) {
		printf("nlsysevent: unknown system for '%s'\n", system);
		return;
	}
	struct nlmsghdr hdr = {};
	if (!nlmsg_reply(&nw, &hdr, sizeof(struct genlmsghdr)))
		return;

	/* XXX also need to send to the "all" group */
	if (nlmsg_get_group_writer(&nw, NLMSG_LARGE, NETLINK_GENERIC, group_id)) {
		struct genlmsghdr *ghdr = nlmsg_reserve_object(&nw, struct genlmsghdr);
		ghdr->version = 0;
		ghdr->cmd = 0;
		ghdr->reserved = 0;
		nlattr_add_string(&nw, NLSE_SYSTEM, system);
		nlattr_add_string(&nw, NLSE_SUBSYSTEM, subsystem);
		nlattr_add_string(&nw, NLSE_TYPE, type);
		if (data != NULL)
			nlattr_add_string(&nw, NLSE_DATA, data);
	}
	nlmsg_end(&nw);
	nlmsg_flush(&nw);
}

static void
nlsysevent_load(void)
{
	devctl_set_notify_hook(sysevent_send);
	ctrl_family_id = genl_register_family(CTRL_FAMILY_NAME, 0, 2, CTRL_ATTR_MAX);
	for (size_t i = 0; i < nitems(sysevents); i++)
		sysevents[i].id = genl_register_group(CTRL_FAMILY_NAME, sysevents[i].name);
	all_group_id = genl_register_group(CTRL_FAMILY_NAME, "all");
}

static void
nlsysevent_unload(void)
{
	devctl_unset_notify_hook();
	genl_unregister_family(CTRL_FAMILY_NAME);
}

static int
nlsysevent_loader(module_t mod __unused, int what, void *priv __unused)
{
	int err = 0;

	switch (what) {
	case MOD_LOAD:
		nlsysevent_load();
		break;
	case MOD_UNLOAD:
		nlsysevent_unload();
		break;
	default:
		err = EOPNOTSUPP;
		break;
	}
	return (err);
}
static moduledata_t nlsysevent_mod = { "nlsysevent", nlsysevent_loader, NULL};

DECLARE_MODULE(nlsysevent, nlsysevent_mod, SI_SUB_PSEUDO, SI_ORDER_ANY);
MODULE_DEPEND(nlsysevent, netlink, 1, 1, 1);
MODULE_VERSION(nlsysevent, 1);
